'\"
'\" Soar man page 
'\"
.so man.macros
.HS output-strings-destination soar
.BS
'\" Note:  do not modify the .SH NAME line immediately below!
.SH NAME
output-strings-destination \- Specify destination of print actions
.SH SYNOPSIS
\fBoutput-strings-destination \fIaction\fR
.BE

.SH DESCRIPTION
.PP
This command redirects the printing of output strings.  This is
useful for applications which need to change where printed results
are placed.  Printed output is normally sent to standard output.
However, if a Graphical User Interface (GUI) is being used in
place of the command line interpreter, then printed output should
appear in the GUI.  GUIs are composed of elements called "widget"s
and printed output would be directed to a "text widget", or procedure. 
.PP
Printed output is normally sent to standard output which is an open file
descriptor in UNIX.  Printed output can be sent to any other 
open file descriptor as well, such as an open file or pipe.  
It is also possible that the 
user is not interested in an agent's printed
output.  In that case, the printed output can be discarded -- which
results in faster processing for the agent as well.  
.PP
Some Soar commands
return results and some print results.  If the user wishes to
have printed results returned so that they can be saved for later
use, then this command can setup the agent to append the printed
output to the normal result.  This capability is also useful when 
sending Soar commands from remote Tcl/Tk interpreters.  By appending
the printed results, all printed characters can be returned to
the remote interpreter rather than going to a screen.
.PP
The printing facility is implemented as a stack, so newly installed
print redirections are in force until popped off the stack.  This is
done to allow easy transient redirection and the restoration of
prior printing contexts after completing a printing task.  Hence,
there are two primary \fIaction\fRs, \fB-push\fR and \fB-pop\fR.
The \fB-pop\fR action takes no additional arguments as it serves
only to pop the print-redirection stack to re-establish the prior
printing context.  The \fB-push\fR action takes the following
additional arguments:
.TP
\fB-append-to-result\fR
Instead of printing, append the output string to the result returned
by a command.  This option is useful for getting all the output when the
command is invoked remotely.
.TP
\fB-channel\fR \fIopen-file-id\fR
Redirect prints to the open channel (file or pipe) denoted
by \fIopen-file-id\fR
.TP
\fB-discard\fR 
Ignore prints (similar to redirecting to /dev/null in the Bourne
shell, sh)
.TP
\fB-procedure\fR \fIproc-name\fR 
Redirect prints to the Tcl procedure named \fIproc-name\fR in the
current interpreter.  Tk widgets are also procedures, so this action
sends output to widget procedures (or any other Tcl procedure).
.TP
\fB-text-widget\fR \fIwidget-name\fR [\fIinterp-name\fR]
Redirect prints to the text widget named \fIwidget-name\fR in the
current interpreter.  If \fIinterp-name\fR is specified, then
prints are redirected to the text widget named \fIwidget-name\fR
in the interpreter named \fIinterp-name\fR.

.SH EXAMPLES
This example redirects prints to the text widget ".text":
.DS
\fBoutput-strings-destination -push -text-widget .text\fR
.DE
.PP

.PP
This example redirects prints to the text widget ".textual" in
the interpreter named \fBcontrol\fR:
.DS
\fBoutput-strings-destination -push -text-widget .textual control\fR
.DE
.PP
This example redirects prints to the open file "stdout":
.DS
\fBoutput-strings-destination -push -file-id stdout\fR
.DE
.PP
This example causes all printing to be supressed:
.DS
\fBoutput-strings-destination -push -discard\fR
.DE
.PP
This example removes the most recently added print destination:
.DS
\fBoutput-strings-destination -pop\fR
.DE

.SH WARNINGS

This command only affects the printing generated by Soar commands such 
as \fBprint\fR and \fBecho\fR.  It does not affect printing done
by the Tcl commands such as \fBputs\fR.

.SH SEE ALSO
\fBecho\fR, \fBprint\fR, \fBlog\fR

